"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TargetLanguage_1 = require("../TargetLanguage");
const Type_1 = require("../Type");
const ConvenienceRenderer_1 = require("../ConvenienceRenderer");
const Naming_1 = require("../Naming");
const Strings_1 = require("../support/Strings");
const Support_1 = require("../support/Support");
const Source_1 = require("../Source");
const TypeUtils_1 = require("../TypeUtils");
const Transformers_1 = require("../Transformers");
const collection_utils_1 = require("collection-utils");
const unicode = require("unicode-properties");
const forbiddenTypeNames = ["True", "False", "None", "Enum", "List", "Dict", "Optional", "Union", "Iterable"];
const forbiddenPropertyNames = [
    "and",
    "as",
    "assert",
    "bool",
    "break",
    "class",
    "continue",
    "datetime",
    "def",
    "del",
    "elif",
    "else",
    "except",
    "finally",
    "float",
    "for",
    "from",
    "global",
    "if",
    "import",
    "in",
    "int",
    "is",
    "lambda",
    "nonlocal",
    "not",
    "or",
    "pass",
    "raise",
    "return",
    "str",
    "try",
    "while",
    "with",
    "yield"
];
class PythonTargetLanguage extends TargetLanguage_1.TargetLanguage {
    getOptions() {
        return [];
    }
    get stringTypeMapping() {
        const mapping = new Map();
        // No Python programmer apparently ever needed to parse ISO date/times.
        // It's only supported in Python 3.7, which isn't out yet.
        const dateTimeType = "string";
        mapping.set("date", dateTimeType);
        mapping.set("time", dateTimeType);
        mapping.set("date-time", dateTimeType);
        // FIXME: Implement transformers
        // mapping.set("integer-string", "integer-string");
        return mapping;
    }
    get supportsUnionsWithBothNumberTypes() {
        return true;
    }
    get supportsOptionalClassProperties() {
        return false;
    }
    needsTransformerForType(t) {
        if (t instanceof Type_1.UnionType) {
            return collection_utils_1.iterableSome(t.members, m => this.needsTransformerForType(m));
        }
        if (t instanceof Type_1.ArrayType) {
            return this.needsTransformerForType(t.items);
        }
        return t.kind !== "string" && Type_1.isPrimitiveStringTypeKind(t.kind);
    }
    makeRenderer(renderContext, _untypedOptionValues) {
        return new JSONPythonRenderer(this, renderContext);
    }
}
exports.PythonTargetLanguage = PythonTargetLanguage;
function isNormalizedStartCharacter(utf16Unit) {
    // FIXME: add Other_ID_Start - https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    // FIXME: also NFKC normalization
    if (utf16Unit === 0x5f)
        return true;
    const category = unicode.getCategory(utf16Unit);
    return ["Lu", "Ll", "Lt", "Lm", "Lo", "Nl"].indexOf(category) >= 0;
}
function isNormalizedPartCharacter(utf16Unit) {
    // FIXME: add Other_ID_Continue - https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    // FIXME: also NFKC normalization
    if (isStartCharacter(utf16Unit))
        return true;
    const category = unicode.getCategory(utf16Unit);
    return ["Mn", "Mc", "Nd", "Pc"].indexOf(category) >= 0;
}
function isStartCharacter(utf16Unit) {
    const s = String.fromCharCode(utf16Unit).normalize("NFKC");
    const l = s.length;
    if (l === 0 || !isNormalizedStartCharacter(s.charCodeAt(0)))
        return false;
    for (let i = 1; i < l; i++) {
        if (!isNormalizedPartCharacter(s.charCodeAt(i)))
            return false;
    }
    return true;
}
function isPartCharacter(utf16Unit) {
    const s = String.fromCharCode(utf16Unit).normalize("NFKC");
    const l = s.length;
    for (let i = 0; i < l; i++) {
        if (!isNormalizedPartCharacter(s.charCodeAt(i)))
            return false;
    }
    return true;
}
const legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function classNameStyle(original) {
    const words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
}
function snakeNameStyle(original, uppercase) {
    const wordStyle = uppercase ? Strings_1.allUpperWordStyle : Strings_1.allLowerWordStyle;
    const words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, wordStyle, wordStyle, wordStyle, wordStyle, "_", isStartCharacter);
}
class PythonRenderer extends ConvenienceRenderer_1.ConvenienceRenderer {
    constructor() {
        super(...arguments);
        this.imports = new Map();
        this.declaredTypes = new Set();
    }
    forbiddenNamesForGlobalNamespace() {
        return forbiddenTypeNames;
    }
    forbiddenForObjectProperties(_, _classNamed) {
        return { names: forbiddenPropertyNames, includeGlobalForbidden: false };
    }
    makeNamedTypeNamer() {
        return Naming_1.funPrefixNamer("type", classNameStyle);
    }
    namerForObjectProperty() {
        return Naming_1.funPrefixNamer("property", s => snakeNameStyle(s, false));
    }
    makeUnionMemberNamer() {
        return null;
    }
    makeEnumCaseNamer() {
        return Naming_1.funPrefixNamer("enum-case", s => snakeNameStyle(s, true));
    }
    get commentLineStart() {
        return "# ";
    }
    emitDescriptionBlock(lines) {
        if (lines.length === 1) {
            this.emitLine('"""', lines[0], '"""');
        }
        else {
            this.emitCommentLines(lines, "", undefined, '"""', '"""');
        }
    }
    get needsTypeDeclarationBeforeUse() {
        return true;
    }
    canBeForwardDeclared(t) {
        const kind = t.kind;
        return kind === "class" || kind === "enum";
    }
    emitBlock(line, f) {
        this.emitLine(line);
        this.indent(f);
    }
    withImport(module, name) {
        collection_utils_1.mapUpdateInto(this.imports, module, s => (s ? collection_utils_1.setUnionInto(s, [name]) : new Set([name])));
        return name;
    }
    withTyping(name) {
        return this.withImport("typing", name);
    }
    namedType(t) {
        const name = this.nameForNamedType(t);
        if (this.declaredTypes.has(t))
            return name;
        return ["'", name, "'"];
    }
    pythonType(t) {
        const actualType = Transformers_1.followTargetType(t);
        return TypeUtils_1.matchType(actualType, _anyType => this.withTyping("Any"), _nullType => "None", _boolType => "bool", _integerType => "int", _doubletype => "float", _stringType => "str", arrayType => [this.withTyping("List"), "[", this.pythonType(arrayType.items), "]"], classType => this.namedType(classType), mapType => [this.withTyping("Dict"), "[str, ", this.pythonType(mapType.values), "]"], enumType => this.namedType(enumType), unionType => {
            const maybeNullable = TypeUtils_1.nullableFromUnion(unionType);
            if (maybeNullable !== null) {
                return [this.withTyping("Optional"), "[", this.pythonType(maybeNullable), "]"];
            }
            const memberTypes = Array.from(unionType.sortedMembers).map(m => this.pythonType(m));
            return [this.withTyping("Union"), "[", collection_utils_1.arrayIntercalate(", ", memberTypes), "]"];
        }, transformedStringType => {
            if (transformedStringType.kind === "date-time") {
                return this.withImport("datetime", "datetime");
            }
            return Support_1.panic(`Transformed type ${transformedStringType.kind} not supported`);
        });
    }
    declarationLine(t) {
        if (t instanceof Type_1.ClassType) {
            return ["class ", this.nameForNamedType(t), ":"];
        }
        if (t instanceof Type_1.EnumType) {
            return ["class ", this.nameForNamedType(t), "(", this.withImport("enum", "Enum"), "):"];
        }
        return Support_1.panic(`Can't declare type ${t.kind}`);
    }
    declareType(t, emitter) {
        this.emitDescription(this.descriptionForType(t));
        this.emitBlock(this.declarationLine(t), () => {
            emitter();
        });
        this.declaredTypes.add(t);
    }
    emitClassMembers(_t) {
        return;
    }
    emitClass(t) {
        this.declareType(t, () => {
            if (t.getProperties().size === 0) {
                this.emitLine("pass");
            }
            else {
                this.forEachClassProperty(t, "none", (name, jsonName, cp) => {
                    this.emitDescription(this.descriptionForClassProperty(t, jsonName));
                    this.emitLine(name, ": ", this.pythonType(cp.type));
                });
                this.ensureBlankLine();
            }
            this.emitClassMembers(t);
        });
    }
    emitEnum(t) {
        this.declareType(t, () => {
            this.forEachEnumCase(t, "none", (name, jsonName) => {
                this.emitLine([name, ' = "', Strings_1.stringEscape(jsonName), '"']);
            });
        });
    }
    emitDeclaration(decl) {
        if (decl.kind === "forward") {
            // We don't need forward declarations yet, since we only generate types.
        }
        else if (decl.kind === "define") {
            const t = decl.type;
            if (t instanceof Type_1.ClassType) {
                this.emitClass(t);
            }
            else if (t instanceof Type_1.EnumType) {
                this.emitEnum(t);
            }
            else if (t instanceof Type_1.UnionType) {
                return;
            }
            else {
                return Support_1.panic(`Cannot declare type ${t.kind}`);
            }
        }
        else {
            return Support_1.assertNever(decl.kind);
        }
    }
    emitImports() {
        this.imports.forEach((names, module) => {
            this.emitLine("from ", module, " import ", Array.from(names).join(", "));
        });
    }
    emitDefaultLeadingComments() {
        this.emitCommentLines(["Only Python >=3.6 right now.  More features coming soon!"]);
    }
    emitSupportCode() {
        return;
    }
    emitClosingCode() {
        return;
    }
    emitSourceStructure(_givenOutputFilename) {
        const declarationLines = this.gatherSource(() => {
            this.forEachDeclaration("interposing", decl => this.emitDeclaration(decl));
        });
        const closingLines = this.gatherSource(() => this.emitClosingCode());
        const supportLines = this.gatherSource(() => this.emitSupportCode());
        if (this.leadingComments !== undefined) {
            this.emitCommentLines(this.leadingComments);
        }
        else {
            this.emitDefaultLeadingComments();
        }
        this.ensureBlankLine();
        this.emitImports();
        this.ensureBlankLine();
        this.emitGatheredSource(supportLines);
        this.ensureBlankLine();
        this.emitGatheredSource(declarationLines);
        this.ensureBlankLine();
        this.emitGatheredSource(closingLines);
    }
}
exports.PythonRenderer = PythonRenderer;
const converterFunctionCodes = {
    none: `def from_none(x):
    assert x is None
    return x`,
    bool: `def from_bool(x):
    assert isinstance(x, bool)
    return x`,
    int: `def from_int(x):
    assert isinstance(x, int) and not isinstance(x, bool)
    return x`,
    "from-float": `def from_float(x):
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)`,
    "to-float": `def to_float(x):
    assert isinstance(x, float)
    return x`,
    str: `def from_str(x):
    assert isinstance(x, str)
    return x`,
    "to-enum": `def to_enum(c, x):
    assert isinstance(x, c)
    return x.value`,
    list: `def from_list(f, x):
    assert isinstance(x, list)
    return [f(y) for y in x]`,
    "to-class": `def to_class(c, x):
    assert isinstance(x, c)
    return x.to_dict()`,
    dict: `def from_dict(f, x):
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }`,
    union: `def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False`,
    "from-datetime": `def from_datetime(x):
    # This is not correct.  Python <3.7 doesn't support ISO date-time
    # parsing in the standard library.  This is a kludge until we have
    # that.
    return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S.%fZ")`,
    "to-datetime": `def to_datetime(x):
    assert isinstance(x, datetime)
    return x.isoformat()`
};
function lambda(x, ...body) {
    return Source_1.multiWord(" ", "lambda", [x, ":"], body);
}
function callFn(fn, ...args) {
    return [Source_1.parenIfNeeded(fn), "(", collection_utils_1.arrayIntercalate(", ", args), ")"];
}
const identityFn = lambda("x", "x");
class JSONPythonRenderer extends PythonRenderer {
    constructor() {
        super(...arguments);
        this._deserializerFunctions = new Set();
        this._converterNamer = Naming_1.funPrefixNamer("converter", s => snakeNameStyle(s, false));
        this._topLevelConverterNames = new Map();
    }
    conv(cf) {
        this._deserializerFunctions.add(cf);
        const name = cf.replace("-", "_");
        if (cf.startsWith("from-") || cf.startsWith("to-"))
            return name;
        return ["from_", name];
    }
    convFn(cf) {
        return Source_1.singleWord(this.conv(cf));
    }
    deserializerFn(t) {
        return TypeUtils_1.matchType(t, _anyType => identityFn, _nullType => this.convFn("none"), _boolType => this.convFn("bool"), _integerType => this.convFn("int"), _doubleType => this.convFn("from-float"), _stringType => this.convFn("str"), arrayType => lambda("x", this.conv("list"), "(", Source_1.parenIfNeeded(this.deserializerFn(arrayType.items)), ", x)"), classType => Source_1.singleWord(this.nameForNamedType(classType)), mapType => lambda("x", this.conv("dict"), "(", Source_1.parenIfNeeded(this.deserializerFn(mapType.values)), ", x)"), enumType => Source_1.singleWord(this.nameForNamedType(enumType)), unionType => {
            const deserializers = Array.from(unionType.members).map(m => this.deserializerFn(m).source);
            return lambda("x", this.conv("union"), "([", collection_utils_1.arrayIntercalate(", ", deserializers), "], x)");
        }, transformedStringType => {
            if (transformedStringType.kind === "date-time") {
                return lambda("x", callFn(this.convFn("to-datetime"), "x"));
            }
            return Support_1.panic(`Transformed type ${transformedStringType.kind} not supported`);
        });
    }
    deserializer(value, t) {
        return callFn(this.deserializerFn(t), value);
    }
    serializerFn(t) {
        return TypeUtils_1.matchType(t, _anyType => identityFn, _nullType => this.convFn("none"), _boolType => this.convFn("bool"), _integerType => this.convFn("int"), _doubleType => this.convFn("to-float"), _stringType => this.convFn("str"), arrayType => lambda("x", this.conv("list"), "(", Source_1.parenIfNeeded(this.serializerFn(arrayType.items)), ", x)"), classType => lambda("x", callFn(this.convFn("to-class"), this.nameForNamedType(classType), "x")), mapType => lambda("x", this.conv("dict"), "(", Source_1.parenIfNeeded(this.serializerFn(mapType.values)), ", x)"), enumType => lambda("x", callFn(this.convFn("to-enum"), this.nameForNamedType(enumType), "x")), unionType => {
            const serializers = Array.from(unionType.members).map(m => this.serializerFn(m).source);
            return Source_1.multiWord("", "lambda x: ", this.conv("union"), "([", collection_utils_1.arrayIntercalate(", ", serializers), "], x)");
        }, transformedStringType => {
            if (transformedStringType.kind === "date-time") {
                return lambda("x", "x.isoformat()");
            }
            return Support_1.panic(`Transformed type ${transformedStringType.kind} not supported`);
        });
    }
    serializer(value, t) {
        return [Source_1.parenIfNeeded(this.serializerFn(t)), "(", value, ")"];
    }
    emitClassMembers(t) {
        this.emitBlock("def __init__(self, obj):", () => {
            this.emitLine("assert isinstance(obj, dict)");
            this.forEachClassProperty(t, "none", (name, jsonName, cp) => {
                const property = ['obj.get("', Strings_1.stringEscape(jsonName), '")'];
                this.emitLine("self.", name, " = ", this.deserializer(property, cp.type));
            });
        });
        this.ensureBlankLine();
        this.emitBlock("def to_dict(self):", () => {
            this.emitLine("result = {}");
            this.forEachClassProperty(t, "none", (name, jsonName, cp) => {
                const property = ["self.", name];
                this.emitLine('result["', Strings_1.stringEscape(jsonName), '"] = ', this.serializer(property, cp.type));
            });
            this.emitLine("return result");
        });
    }
    emitSupportCode() {
        for (const df of this._deserializerFunctions) {
            this.emitMultiline(converterFunctionCodes[df]);
            this.ensureBlankLine();
        }
    }
    makeTopLevelDependencyNames(_t, topLevelName) {
        const fromDict = new Naming_1.DependencyName(this._converterNamer, ConvenienceRenderer_1.topLevelNameOrder, l => `${l(topLevelName)}_from_dict`);
        const toDict = new Naming_1.DependencyName(this._converterNamer, ConvenienceRenderer_1.topLevelNameOrder, l => `${l(topLevelName)}_to_dict`);
        this._topLevelConverterNames.set(topLevelName, { fromDict, toDict });
        return [fromDict, toDict];
    }
    emitClosingCode() {
        this.forEachTopLevel("interposing", (t, name) => {
            const { fromDict, toDict } = Support_1.defined(this._topLevelConverterNames.get(name));
            const pythonType = this.pythonType(t);
            this.emitBlock(["def ", fromDict, "(s: str) -> ", pythonType, ":"], () => {
                this.emitLine("return ", this.deserializer("s", t));
            });
            this.ensureBlankLine();
            this.emitBlock(["def ", toDict, "(x: ", pythonType, ") -> str:"], () => {
                this.emitLine("return ", this.serializer("x", t));
            });
        });
    }
}
exports.JSONPythonRenderer = JSONPythonRenderer;
